<?php

namespace BeycanPress\CryptoPay\Donate\DonateBox;

use \BeycanPress\CryptoPay\Donate\Lang;
use \BeycanPress\CryptoPay\Donate\Settings;
use \BeycanPress\CryptoPay\Donate\Models\Address;
use \BeycanPress\CryptoPay\Donate\Models\Transaction;
use \BeycanPress\CryptoPay\Donate\PluginHero\Helpers;

class DonateBox
{
    use Helpers;

    /**
     * @var object
     */
    private $api;

    /**
     * @var string
     */
    private $receiver;

    public function __construct()
    {
        add_action('init', function(){
            $this->api = new DonateBoxApi();
            add_shortcode('cryptopay-donate-box', array($this, 'init'));
            add_shortcode('cryptopay-donate-box-list', array($this, 'list'));
        });
    }

    public function init($atts)
    {
        extract(shortcode_atts([
            'receiver' => null
        ], $atts));

        if ($receiver && !$this->isAddress($receiver)) {
            return esc_html__('Invalid receiver address', 'cryptopay');
        }

        $this->receiver = !is_null($receiver) ? trim($receiver) : $this->setting('payment_address');

        $this->loadScripts();

        return $this->view('donate-box');
    }

    public function list()
    {
        $address = new Address();
        $transaction = new Transaction();

        if (!is_user_logged_in()) {
            return esc_html__('Only logged in users can access this page.', 'cryptopay');
        }
        
        $address = $address->findOneBy([
            'user_id' => get_current_user_id()
        ]);

        if (!$address) {
            return esc_html__('No address matching the user found.', 'cryptopay');
        }
        
        $transactions = $transaction->findBy([
            'receiver' => $address->address
        ], ['created_at', 'DESC']);

        return $this->view('donations-for-user', compact('transactions'));
    }

    /**
     * javascript ve css dosyalarını dahil eder
     * @return void
     */
    public function loadScripts()
    { 
        $this->addScript('js/multi-chain.min.js');
        $this->addScript('cryptopay/js/chunk-vendors.js');
        $this->addScript('cryptopay/js/app.js');
        $this->addStyle('cryptopay/css/app.css');
        $key = $this->addScript('js/main.js');
        wp_localize_script($key, 'CryptoPayDonate', $this->jsData());
    }
    
    /**
     * Javascript'e gönderilecek dinamik veriyi hazırlar
     * @return array
     */
    private function jsData()
    {
        return [
            'mode' => 'donate',
            'lang' => Lang::get(),
            'acceptedChains' => $this->prepareChains(),
            'acceptedWallets' => Settings::getAcceptedWallets(),
            'apiUrl' => $this->api->getUrl(),
            'imagesUrl' => $this->pluginUrl . 'assets/images/',
            'receiver' => $this->receiver, 
            'testnets' => (bool) $this->setting('testnets'),
            'infuraId' => $this->setting('infuraProjectId')
        ];
    }

    /**
     * Zincirleri cryptopay-core'un anlayacağı şekilde hazırlar
     * @return array
     */
    public function prepareChains()
    {
        $preparedChains = [];
        $acceptedChains = $this->setting('accepted_chains');

        if (!$acceptedChains) return $preparedChains;

        foreach ($acceptedChains as $key => $chain) {

            // Active/Passive control
            if ($chain['active'] != '1') continue;

            $id = intval($chain['id']);
            $hexId = '0x' . dechex($id);

            $currencies = [];

            if ($chain['native_currency']['active'] == '1') {
                unset($chain['native_currency']['active']);
                $chain['native_currency']['address'] = trim(strtoupper($chain['native_currency']['symbol']));
                $chain['native_currency']['symbol'] = trim(strtoupper($chain['native_currency']['symbol']));
                $chain['native_currency']['decimals'] = (int) $chain['native_currency']['decimals'];
                $currencies[$chain['native_currency']['symbol']] = $chain['native_currency'];
            }
            
            foreach ($chain['currencies'] as $key => $curency) {
                if ($curency['active'] == '1') {
                    unset($curency['active']);
                    $curency['symbol'] = trim(strtoupper($curency['symbol']));
                    $curency['address'] = trim(strtoupper($curency['address']));
                    $currencies[$curency['address']] = $curency;
                }
            }

            $preparedChains[$hexId] = [
                'id' => $id,
                'hexId' => $hexId,
                'name' => $chain['name'],
                'rpcUrl' => $chain['rpc_url'],
                'explorerUrl' => $chain['explorer_url'],
                'nativeCurrency' => $chain['native_currency'],
                'currencies' => $currencies
            ];
        }
        
        return $preparedChains;
    }

    /**
     * @param string $key
     * @return bool
     */
    private function isAddress(string $address) : bool
    {
        if (!is_string($address)) {
            return false;
        }
        return (preg_match('/^0x[a-fA-F0-9]{40}$/', $address) >= 1);
    }
}